#include "myfndt.h"
#include "frontln.h"
#include "docum.h"
#include "tl.h"
#include "reporter.h"

const ulong CFrontLine::FOUninit = 0xffffffff;
const ulong CFrontLine::FOEmpty = 0xfffffffe;

CFrontLine::CFrontLine(const DString &TempDir,
    ulong aTotalDocCount,
	int TimeoutSec,
	int &aBreakFlag):
    TotalDocCount(aTotalDocCount),
    Timeout(TimeoutSec),
    FirstOcc(FOUninit),
    Front(4, 12),
    Unknown(4, 12),
    Cache(10, 20),
    Storage(new CULongHandler(), // DDocument depends on this, too
	   TempDir,
	   new CDatumFactoryTemplate<CAnswer>,
	   TempDir,
	   TempDir,
	   IT_TEMP | IT_UPDATE),
	BreakFlag(aBreakFlag),
	Broken(0),
	FreeIndex(0)
{  
}

CFrontLine::~CFrontLine()
{
}

int CFrontLine::AddVariation(CFrontItem *InItem)
{
    CObjPtr<CFrontItem> Item = InItem;
    Assert(!IsInitialized());
    Assert(InItem->GetResidentIndex() <= FreeIndex);

    if (InItem->GetResidentIndex() == FreeIndex)
	   ++FreeIndex;

    int Index = Front.Find(InItem); 
    if (Index == INT_MAX)
	   Front.Add(Item.Drop());
    else
    {   CFrontItem *Resident = Front[Index];
	   Resident->Indices.Add(InItem->GetResidentIndex());
    }

    if (BreakFlag)
    {   Broken = 1;
	   return 1;
    }
    else
	   return 0;
}

int CFrontLine::GetFirstOcc(ulong &GlobOcc)
{
    if (!IsInitialized())
	   Init();
    if (FirstOcc == FOEmpty)
	   return 1;

    GlobOcc = FirstOcc;
    return 0;
}

void CFrontLine::AddCachedDoc(CLineDoc *InDoc)
{
    CObjPtr<CLineDoc> Doc = InDoc;
    Assert(IsInitialized());
    
    if (Cache.Find(Doc) == INT_MAX)
	   Cache.Add(Doc.Drop());
}

int CFrontLine::SkipDoc(ulong DocNum,
    ulong &GlobOcc)
{
    Assert(IsInitialized());

    CLineDoc Test(DocNum, CLineDoc::TestFirstWord);
    int Index = Cache.Find(&Test);
    if (Index == INT_MAX)
	   return 2;

    return SkipDocIntern(Cache[Index], GlobOcc);
}

int CFrontLine::SkipDoc(ulong &GlobOcc)
{
    Assert(IsInitialized());

    CLineDoc Test(CLineDoc::TestDocNum, GlobOcc);
    DISArray<CLineDoc>::DIterator Itr = lower_bound(Cache.begin(), Cache.end(),
            DDumbPtr<CLineDoc>(&Test));
    if (Itr == Cache.end())
	   return 2;

    CLineDoc *Found = (*Itr).p;
    if (GlobOcc >= Found->GetEndOcc())
	   return 2;

    return SkipDocIntern(Found, GlobOcc);
}

DDocument *CFrontLine::NewDoc(ulong DocNum)
{
    Assert(IsInitialized());

    CLineDoc Test(DocNum, CLineDoc::TestFirstWord);
    int Index = Cache.Find(&Test);
    if (Index == INT_MAX)
	   return NULL;

    return NewDocIntern(Cache[Index]);
}

DISArray<DString> &CFrontLine::GetUnknown()
{
    if (!IsInitialized())
	   Init();
    return Unknown;
}

void CFrontLine::Init()
{   ulong Cur;

    int Count = Front.GetCount();
    FirstOcc = FOEmpty;
    for (int i = 0; i < Count; i++)
    {   if (!Front[i]->GetCurOcc(Cur))
	   {   if (FirstOcc == FOEmpty || Cur < FirstOcc)
	     	 FirstOcc = Cur;
	   }
	   else
		  Unknown.Add(new DString(Front[i]->WordVariation));
    }
}

int CFrontLine::SkipDocIntern(CLineDoc *LineDoc, ulong &GlobOcc)
{   int i;
    ulong Occ;
    ulong DummyOcc;
   
    int Count = Front.GetCount();

    ulong HighBound = LineDoc->GetEndOcc();
    for (i = 0; i < Count; i++)
    {   CFrontItem *FrontItem = Front[i];
		Assert(FrontItem->GetCurOcc(DummyOcc) || 
			DummyOcc >= LineDoc->GetBeginOcc());
		while (!FrontItem->GetCurOcc(Occ) && Occ < HighBound)
        {   
			FrontItem->Iterate();
			if (BreakFlag || Timeout.IsSignaled())
			{	Broken = 1;
				return 1;
			}
		}
    }		  

    GlobOcc = FOEmpty;
    for (i = 0; i < Count; i++)
    {   CFrontItem *FrontItem = Front[i];
        if (!FrontItem->GetCurOcc(Occ))
		{	if (GlobOcc == FOEmpty || Occ < GlobOcc)
				GlobOcc = Occ;
		}
		if (BreakFlag || Timeout.IsSignaled())
		{	Broken = 1;
			return 1;
		}
    }

    if (ShrinkCache(LineDoc))
		return 1;

    return GlobOcc != FOEmpty ? 0 : 1;
}

DDocument *CFrontLine::NewDocIntern(CLineDoc *LineDoc)
{   ulong Occ;
    ulong DummyOcc;

    CNewKey Key(CULongHandler::NewKey(LineDoc->DocNum));
    CAnswer *Answer = static_cast<CAnswer *>(Storage.LockAlways(Key));
    CObjPtr<DDocument> Document = new DDocument(Answer, 
		Storage, 
        LineDoc->DocNum,
		0,
        1);

    int Count = Front.GetCount();
    ulong LowBound = LineDoc->GetBeginOcc();
    ulong HighBound = LineDoc->GetEndOcc();

    for (int i = 0; i < Count; i++)
    {   CFrontItem *FrontItem = Front[i];
        Assert(FrontItem->GetCurOcc(DummyOcc) || 
            DummyOcc >= LowBound);
        while (!FrontItem->GetCurOcc(Occ) && Occ < HighBound)
        {   AddIndices(Answer, FrontItem->Indices, Occ - LowBound);
            FrontItem->Iterate();
            if (BreakFlag || Timeout.IsSignaled())
            {	Broken = 1;
                return NULL;
            }
        }
    }

    if (ShrinkCache(LineDoc))
		return NULL;

    return Document.Drop();
}

// this would be more effective were the cache ordered _backwards_
int CFrontLine::ShrinkCache(CLineDoc *LineDoc)
{
    CLineDoc *Doomed = Cache[0];
    while (!(*Doomed == *LineDoc))
    {	Cache.DetachByIndex(0);
		delete Doomed;
		Doomed = Cache[0];
		if (BreakFlag)
		{	Broken = 1;
			return 1;
		}
    }
	return 0;
}

void CFrontLine::AddIndices(CAnswer *Answer, 
    DArray<ushort> &Indices, 
    ulong LocalOccNum)
{
    DArray<ushort>::DIterator Itr = Indices.InitIterator();
    while (Indices.IsDereferenceable(Itr))
    {   Answer->Add(*Itr, LocalOccNum);
	   ++Itr;
    }
}
