#include "myfndt.h"
#include "softop.h"
#include "docwrap.h"
#include "result.h"
#include "maskans.h"
#include "lineans.h"
#include "relevanc.h"
#include "wordmask.h"
#include "filter.h"
#include "gridans.h"

#include "osoftop.h"
#include "oswild.h"
#include "ooutform.h"

#include <string.h>
#include <stdlib.h>
#include <math.h>

class CSoftOperator::CRelAlgorithm
{
private:
    int Relevant;
    CMaskingAnswer *MaskingAnswer;

    long WordCount;
    int TryCount;
    long HitCount;
    int UniqueCount;
    long StartPos;
    long RepreLen;
    int SeqCount;

public:
	CRelAlgorithm(CMaskingAnswer *InMaskingAnswer,
        int aLineWordCount);
	virtual ~CRelAlgorithm();

	CRelevance *NewRelevance(); // returns NULL if doc not relevant

private:
    double GetFocus();
    double GetCorrelation();
    double GetProminence();
    double GetLocality();
    double GetOrder();

    // these methods are called from constructor on only partially initialized
    // object
    short *AllocSums(); 
    void GetUniqueCount(const short *Sums);
    long GetStartPos();
    long GetRepreLen(const short *Sums);
    long GetRepreFromPos(const short *Sums,
	   long Itr);
    int GetSeqCount();
};

CSoftOperator::CRelAlgorithm::CRelAlgorithm(CMaskingAnswer *InMaskingAnswer,
    int aLineWordCount):
    Relevant(0),
    MaskingAnswer(InMaskingAnswer),
    WordCount(InMaskingAnswer->GetVisibleWordsInDocCount()),
    TryCount(aLineWordCount),
    HitCount(0),
    UniqueCount(0),
    StartPos(0),
    RepreLen(0),
    SeqCount(0)
{
    Assert(TryCount > 0);
    if (!WordCount) // can happen for restricted operators
        return;

    CMemPtr<short> Sums = AllocSums();
    if (!Sums)
        return;

    Assert(HitCount > 0);
    GetUniqueCount((short *)Sums);
    Assert(UniqueCount > 0);
    StartPos = GetStartPos();
    RepreLen = GetRepreLen((short *)Sums);
    SeqCount = GetSeqCount();

    Relevant = 1;
}

CSoftOperator::CRelAlgorithm::~CRelAlgorithm()
{
    delete MaskingAnswer;
}

double CSoftOperator::CRelAlgorithm::GetFocus()
{
	return sqrt(sqrt((double)HitCount / WordCount));
		// sqrt may be slow
}

double CSoftOperator::CRelAlgorithm::GetCorrelation()
{
	return (double)UniqueCount / TryCount;
}

double CSoftOperator::CRelAlgorithm::GetProminence()
{
	return 1 - (double)StartPos / WordCount;
}

double CSoftOperator::CRelAlgorithm::GetLocality()
{
	return (1 - (double)RepreLen / WordCount) *
		(double)UniqueCount / TryCount;
}

double CSoftOperator::CRelAlgorithm::GetOrder()
{
	return (double)SeqCount / UniqueCount;
}

CRelevance *CSoftOperator::CRelAlgorithm::NewRelevance()
{
    if (!Relevant)
        return NULL;

    return new CRelevance(GetFocus(),
	   GetCorrelation(),
	   GetProminence(),
	   GetLocality(),
	   GetOrder());
}

short *CSoftOperator::CRelAlgorithm::AllocSums()
{
    CMemPtr<short> Sums(TryCount * sizeof(short));
    for (int i = 0; i < TryCount; i++)
	   *((short *)Sums + i) = 0;

    long Itr = MaskingAnswer->InitIterator();
    if (!Itr)
	   return NULL;

    while (Itr)
    {   ushort WordIndex = MaskingAnswer->GetConstrictedIndex(Itr);
	   ++HitCount;
	   Assert(WordIndex < TryCount);
	   (*((short *)Sums + WordIndex))++;
    }

    return Sums.Drop();
}

void CSoftOperator::CRelAlgorithm::GetUniqueCount(const short *Sums)
{
    UniqueCount = 0;
    for (int i = 0; i < TryCount; i++)
    {   if (Sums[i] > 0)
		  UniqueCount++;
    }
}

long CSoftOperator::CRelAlgorithm::GetStartPos()
{
    long Itr = MaskingAnswer->InitIterator();
    Assert(Itr != 0);
    return MaskingAnswer->GetConstrictedOcc(Itr);
}

long CSoftOperator::CRelAlgorithm::GetRepreLen(const short *Sums)
{
    CMemPtr<short> ChSums(TryCount * sizeof(short));
    memcpy((short *)ChSums, Sums, TryCount * sizeof(short));

    long Itr = MaskingAnswer->InitIterator();
    Assert(Itr != 0);
    long Len = GetRepreFromPos((short *)ChSums,
	   Itr);
    ushort WIndex = MaskingAnswer->GetConstrictedIndex(Itr);
    while (Itr)
    {   // representation can be only in passage which has
	   // all hits found in document
	   Assert(*((short *)ChSums + WIndex) > 0);
	   (*((short *)ChSums + WIndex))--;
	   if (!(*((short *)ChSums + WIndex)))
		  break;

	   long TestLen = GetRepreFromPos((short *)ChSums,
		  Itr);
	   WIndex = MaskingAnswer->GetConstrictedIndex(Itr);

	   if (TestLen < Len)
		  Len = TestLen;
    }

    return Len;
}

long CSoftOperator::CRelAlgorithm::GetRepreFromPos(const short *Sums,
    long Itr)
{   
    long Len = 0;
    long OrigItr = Itr;
    Assert(UniqueCount > 0);
    int aUniqueCount = UniqueCount;
    ulong FirstOcc = 0;

    CMemPtr<short> ChSums(TryCount * sizeof(short));
    memcpy((short *)ChSums, Sums, TryCount * sizeof(short));

    int First = 1;
    while (Itr)
    {   CAnswer::COcc Hit = MaskingAnswer->GetHit(Itr);
	   if (First)
	   {	   First = 0;
		   FirstOcc = MaskingAnswer->GetConstrictedOcc(Hit);
	   }
	   ushort WIndex = MaskingAnswer->GetConstrictedIndex(Hit);
	   Assert(WIndex < TryCount);
	   if (*((short *)ChSums + WIndex) > 0)
	   {	   *((short *)ChSums + WIndex) = 0;
		   aUniqueCount--;
		   if (!aUniqueCount)
		   {   ulong Occ = MaskingAnswer->GetConstrictedOcc(Hit);
               Assert(Occ >= FirstOcc);
			   Len = Occ - FirstOcc; // can be 0
			   goto EXIT;
		   }
	    }
	}
	Assert(0);

EXIT:
    MaskingAnswer->ResetIterator(OrigItr);
    Assert(Len >= 0); 
    return Len;
}

int CSoftOperator::CRelAlgorithm::GetSeqCount()
{
    long Itr = MaskingAnswer->InitIterator();
    Assert(Itr != 0);
    int CurSeqCount = 1;
    int MaxSeqCount = 1;
    ushort CurIndex = MaskingAnswer->GetConstrictedIndex(Itr);
    while (Itr)
    {   ushort WIndex = MaskingAnswer->GetConstrictedIndex(Itr);
	   if (WIndex <= CurIndex) // words in sequence cannot repeat
	   {   // start new sequence
		  if (CurSeqCount > MaxSeqCount)
			 MaxSeqCount = CurSeqCount;
		  CurSeqCount = 1;
	   }
	   else
	   {   // continue sequence
		  CurSeqCount++;
	   }
	   CurIndex = WIndex;
    }
    if (CurSeqCount > MaxSeqCount)
	   MaxSeqCount = CurSeqCount;
    return MaxSeqCount;
}

void CSoftOperator::SetUnoClass(CUnivOptions &Options, 
    const CStringListValue &NamedSoftOps)
{   // must add wild keys - user cannot set tamed ones
    static const char *IntAttr[] = {O_SHOWRELWILD};
    static const char *DoubleAttr[] = {O_WILDFOCUSWEIGHT, O_WILDCORRELATIONWEIGHT,
            O_WILDPROMINENCEWEIGHT, O_WILDLOCALITYWEIGHT, O_WILDORDERWEIGHT,
		  O_RELEVANCETHRESHOLD};

    Options.COptions<int>::SetUnoClass(NamedSoftOps,
        IntAttr,
	   sizeof(IntAttr) / sizeof(IntAttr[0]));
    Options.COptions<double>::SetUnoClass(NamedSoftOps,
        DoubleAttr,
	   sizeof(DoubleAttr) / sizeof(DoubleAttr[0]));
}

CSoftOperator::CSoftOperator(const DString &Name, 
    CSubdocFilter *InFilter, 
    CUnivOptions &Options):
    Filter(InFilter)
{
    // must use wild keys - tamed ones are not there for user-named objects
    Init(Options.COptions<double>::GetUno(Name, O_WILDFOCUSWEIGHT),
	   Options.COptions<double>::GetUno(Name, O_WILDCORRELATIONWEIGHT),
	   Options.COptions<double>::GetUno(Name, O_WILDPROMINENCEWEIGHT),
	   Options.COptions<double>::GetUno(Name, O_WILDLOCALITYWEIGHT),
	   Options.COptions<double>::GetUno(Name, O_WILDORDERWEIGHT),
	   Options.COptions<double>::GetUno(Name, O_RELEVANCETHRESHOLD),
	   Options.COptions<int>::GetUno(Name, O_SHOWRELWILD));
}

CSoftOperator::CSoftOperator(double aFocusWeight,
    double aCorrelationWeight,
    double aProminenceWeight,
    double aLocalityWeight,
    double aOrderWeight,
    double aRelevanceThreshold,
    int aKeepRelevance):
    Filter(NULL)
{
    Init(aFocusWeight, aCorrelationWeight, aProminenceWeight, aLocalityWeight, 
            aOrderWeight, aRelevanceThreshold, aKeepRelevance);
}

void CSoftOperator::Init(double aFocusWeight,
    double aCorrelationWeight,
    double aProminenceWeight,
    double aLocalityWeight,
    double aOrderWeight,
    double aRelevanceThreshold,
    int aKeepRelevance)
{
    RelevanceThreshold = aRelevanceThreshold;
    KeepRelevance = aKeepRelevance;
    Weight = (double *)malloc(CRelevance::FeatureCount * sizeof(double));
    if (!Weight)
	   throw DMemoryException(CRelevance::FeatureCount * sizeof(double));

    Weight[CRelevance::Focus] = aFocusWeight;
    Weight[CRelevance::Correlation] = aCorrelationWeight;
    Weight[CRelevance::Prominence] = aProminenceWeight;
    Weight[CRelevance::Locality] = aLocalityWeight;
    Weight[CRelevance::Order] = aOrderWeight;
}

CSoftOperator::~CSoftOperator()
{
    delete Filter;
	free(Weight);
}

void CSoftOperator::AddRelevant(CResult *Result, 
    CDocWrapper *DocWrapper, 
    CWordMask &Mask)
{
    CMaskingAnswer *Answer = NewMaskingAnswer(DocWrapper, Mask);
    if (!Answer)
        return;

    CRelAlgorithm RelAlgorithm(Answer, Mask.GetCount());
    CObjPtr<CRelevance> Relevance = RelAlgorithm.NewRelevance();
    if (!Relevance)
	   return;

    double RelRes = GetRelRes(Relevance);

    if (RelRes >= RelevanceThreshold)
	   Result->Add(DocWrapper->GetDoc(),
		  RelRes,
		  KeepRelevance ? Relevance.Drop() : 0);
}

CMaskingAnswer *CSoftOperator::NewMaskingAnswer(CDocWrapper *DocWrapper, 
    CWordMask &Mask)
{
    if (!Filter)
        return new CLineAnswer(DocWrapper->GetWordCount(),
            *(DocWrapper->GetAnswer()), 
            Mask);

    if (DocWrapper->GetSubDesc())
        return new CGridAnswer(DocWrapper->GetWordCount(),
            *(DocWrapper->GetAnswer()), 
            Mask,
            *Filter,
            *(DocWrapper->GetSubDesc()));
    
    // this is a special case: user wants to search for
    // subdocs (i. e. documents generally have them), but
    // this document contains only plain text
    if (!Filter->Has(0))
        return NULL; // doc is not relevant
    else
        return new CLineAnswer(DocWrapper->GetWordCount(),
            *(DocWrapper->GetAnswer()), 
            Mask);
}

int CSoftOperator::IsRelevant(CDocWrapper *DocWrapper, 
    CWordMask &Mask, 
    double &Relevance)
{
    CMaskingAnswer *Answer = NewMaskingAnswer(DocWrapper, Mask);
    if (!Answer)
        return 0;

    CRelAlgorithm RelAlgorithm(Answer, Mask.GetCount());
    CObjPtr<CRelevance> Rel = RelAlgorithm.NewRelevance();
    if (!Rel)
	   return 0;

    Relevance = GetRelRes(Rel);
    if (Relevance >= RelevanceThreshold)
	   return 1;
    else
	   return 0;
}

double CSoftOperator::GetRelRes(CRelevance *Relevance)
{
    double RelRes = 0;
    for (CRelevance::EFeature Feature = (CRelevance::EFeature)0;
	   Feature < CRelevance::FeatureCount;
	   Feature = (CRelevance::EFeature)(Feature + 1))
	   RelRes += Relevance->GetMetric(Feature) * Weight[Feature];
    Assert(-0.01 < RelRes && RelRes < 1.01);
    return RelRes;
}
