#include "myfndt.h"
#include "storag.h"
#include "datum.h"
#include "assoc.h"

CStorage::CStorage(CKeyHandler *aKeyHandler,
	const DString &aKeyFile,
	CDatumFactory *aFactory,
	const DString &aFreeListFile,
	const DString &aDataFile,
	short Opt,
	int RemLimit,
	int RemDelta,
	int CacheLimit):
	KeyStorage(aKeyHandler, aKeyFile, Opt),
	DataStorage(aFreeListFile, aDataFile, Opt),
	Factory(aFactory),
	// key handler is owned (i. e. deleted) by CBook
	KeyHandlerAlias(aKeyHandler),
	Remembered(RemLimit, RemDelta, aKeyHandler),
	Cache(CacheLimit, aKeyHandler)
{
}

CStorage::~CStorage()
{
	Assert(Remembered.IsEmpty());
		// do not leave dangling pointers
	delete Factory;
}

CDatum *CStorage::Lock(KEY Key)
{
	CAssoc *Assoc = GetIntern(Key);
	if (!Assoc)
		return NULL;
	return LockIntern(Assoc);
}

CDatum *CStorage::LockAlways(KEY Key)
{
	CAssoc *Assoc = GetIntern(Key);
	if (!Assoc)
	{	Assoc = new CAssoc(KeyHandlerAlias,
			Key,
			Factory->Create(this, 0));
		Remembered.Add(Assoc);
	}
	return LockIntern(Assoc);
}

void CStorage::Unlock(CDatum *Datum)
{
	CAssoc *Assoc = Remembered.Get(Datum);
	Assert(Assoc != NULL);
	if (Assoc->DecLockCount())
	{	Remembered.Detach(Assoc);
		CObjPtr<CAssoc> OldAssoc = Cache.Add(Assoc);
		if (OldAssoc)
			Store(OldAssoc);
	}
}

KEY CStorage::NewFirstKey()
{
	CLockedKey StoredKey(KeyStorage.LockFirstKey(), &KeyStorage);
	KEY Key1 = GetExtremKey(StoredKey,
		Remembered.GetFirstKey(),
		1);
	KEY Key2 = GetExtremKey(Key1,
		Cache.GetFirstKey(),
		1);
	if (!Key2)
		return NULL;
	return KeyHandlerAlias->NewCopy(Key2);
}

KEY CStorage::NewLastKey()
{
	CLockedKey StoredKey(KeyStorage.LockLastKey(), &KeyStorage);
	KEY Key1 = GetExtremKey(StoredKey,
		Remembered.GetLastKey(),
		0);
	KEY Key2 = GetExtremKey(Key1,
		Cache.GetLastKey(),
		0);
	if (!Key2)
		return NULL;
	return KeyHandlerAlias->NewCopy(Key2);
}

KEY CStorage::NewNextKey(KEY Key)
{
	CLockedKey StoredKey(LockGreaterStoredKey(Key), &KeyStorage);
	KEY Key1 = GetExtremKey(StoredKey,
		GetGreaterRememberedKey(Key),
		1);
	KEY Key2 = GetExtremKey(Key1,
		GetGreaterCachedKey(Key),
		1);
	if (!Key2)
		return NULL;
	return KeyHandlerAlias->NewCopy(Key2);
}

short CStorage::Del(KEY Key)
{
     // it's illegal to delete locked key
     Assert(Remembered.Get(Key) == NULL);
	CAssoc *Assoc = GetInternUnlocked(Key);
	if (!Assoc)
		return 1;

	CDatum *Datum = Assoc->GetDatum();
	Datum->Remove();
	KeyStorage.Del(Key);
		// key might have been in cache, not in key storage;
		// Del then returns 1
	Remembered.Detach(Assoc);
	delete Assoc;
	return 0;
}

void CStorage::StoreAll()
{
	Remembered.StoreAll(this);
	Cache.StoreAll(this);
	KeyStorage.StoreAll();
	DataStorage.StoreAll();
}

CAssoc *CStorage::GetIntern(KEY Key)
{	
	CAssoc *Assoc = Remembered.Get(Key);
	if (!Assoc)
	    return GetInternUnlocked(Key);
	return Assoc;
}

int CStorage::IsKnown(KEY Key)
{	
    if (Remembered.Get(Key) != NULL)
	   return 1;
	
    if (Cache.IsKnown(Key))
	   return 1;

    LONG DummyTicket;
    return !(KeyStorage.GetData(Key, DummyTicket));
}

CAssoc *CStorage::GetInternUnlocked(KEY Key)
{   LONG Ticket;
	
    CAssoc *Assoc = Cache.Yank(Key);
    if (!Assoc)
    {   if (KeyStorage.GetData(Key, Ticket))
	       return NULL;
	   Assoc = new CAssoc(KeyHandlerAlias,
		  Key,
		  Factory->Create(this, Ticket));
    }
    Remembered.Add(Assoc);
    return Assoc;
}

CDatum *CStorage::LockIntern(CAssoc *Assoc)
{
	Assoc->IncLockCount();
	return Assoc->GetDatum();
}

void CStorage::Store(CAssoc *Assoc)
{
	CDatum *Datum = Assoc->GetDatum();
	long Ticket = Datum->Finish();
	if (!Ticket)
		return;

	KEY Key = Assoc->GetKey();
	if (KeyStorage.ChangeData(Key, Ticket))
	{	// key does not yet exist
		short KnownRet = KeyStorage.Add(Key, Ticket);
		Assert(!KnownRet);
	}
}

KEY CStorage::GetExtremKey(KEY Key1, KEY Key2, short Min)
{
	if (!Key1)
		return Key2;
	if (!Key2)
		return Key1;
	short CmpRes = KeyHandlerAlias->Compare(Key1, Key2);
	if ((Min && CmpRes < 0) ||
		(!Min && CmpRes > 0))
		return Key1;
	else
		return Key2;
}

KEY CStorage::LockGreaterStoredKey(KEY Key)
{
	CLockedKey StoredKey(KeyStorage.LockKeyAfter(Key), &KeyStorage);
	if (!StoredKey)
		return NULL;
	if (!KeyHandlerAlias->Compare(Key, StoredKey))
		return KeyStorage.LockNextKey(StoredKey);
	else
		return StoredKey.Drop();
}

KEY CStorage::GetGreaterRememberedKey(KEY Key)
{
	KEY RemKey = Remembered.GetKeyAfter(Key);
	if (!RemKey)
		return NULL;
	if (!KeyHandlerAlias->Compare(Key, RemKey))
		return Remembered.GetNextKey(RemKey);
	else
		return RemKey;
}

KEY CStorage::GetGreaterCachedKey(KEY Key)
{
	KEY CachedKey = Cache.GetKeyAfter(Key);
	if (!CachedKey)
		return NULL;
	if (!KeyHandlerAlias->Compare(Key, CachedKey))
		return Cache.GetNextKey(CachedKey);
	else
		return CachedKey;
}

/*#if defined(MEMCHECK)
int CStorage::GetRemCount()
{
	return Remembered.GetCount();
}

int CStorage::GetCacheCount()
{
	return Cache.GetCount();
}
#endif*/

