#include "myfndt.h"
#include "cache.h"
#include "lockptr.h"
#include "tltree.h"
#include "factory.h"

CCacheItem::CCacheItem(CSortedNode *aNode, long aPrevOffs, long aNextOffs):
	Node(aNode), PrevOffs(aPrevOffs), NextOffs(aNextOffs)
{
}

CCacheItem::~CCacheItem()
{
	delete Node;
}

#if defined(NOCPCONSTRUCT)
CCacheItem::operator == (const CCacheItem *Other)
{
	return GetOffs() == Other->GetOffs();
}

CCacheItem::operator < (const CCacheItem *Other)
{
	return GetOffs() < Other->GetOffs();
}

CCacheItem::operator > (const CCacheItem *Other)
{
	return GetOffs() > Other->GetOffs();
}
#endif

const int CCache::Limit = 6;
	/* profile 13. 9.: raising from 3 to 6 made _no_ improvement
		on 1 file, 780 884 bytes (100 minutes) */
	/* profile 10. 5. 1997 (grazer compiled by DJGPP): the three most common words
	in VIRUS-L
		3	25 min
		6	8 min
		10	8 min */

CCache::CCache(CTree *aTree, int aLimit):
	CacheArr(aLimit, 0),
	FirstOffs(NOOFFS),
	LastOffs(NOOFFS),
	Tree(aTree)
{
}

void CCache::Add(CSortedNode *Node)
{
	if (CacheArr.IsFull())
	{	Assert(LastOffs != NOOFFS);
#if defined(DEBUG_duplicate)
		Tree->CheckDupl();
#endif
		Del(LastOffs);
	}

	long OldFirstOffs = FirstOffs;
	CCacheItem *Item = new CCacheItem(Node,
		NOOFFS,
		OldFirstOffs);

	CacheArr.Add(Item);
	FirstOffs = Item->GetOffs();

	if (LastOffs != NOOFFS)
	{	ushort OldFirstIdx = FindIndex(OldFirstOffs);
		Assert(OldFirstIdx != NOIDX);
		CacheArr[OldFirstIdx]->PrevOffs = FirstOffs;
	}
	else
		LastOffs = FirstOffs;

#if defined(DEBUG_CCache)
	Check();
#endif
}

CSortedNode *CCache::Yank(long Offs)
{
	ushort FindIdx = FindIndex(Offs);
	if (FindIdx == NOIDX)
		return NULL;
	CObjPtr<CCacheItem> OrigItem = Detach(Offs, FindIdx);
	CSortedNode *OrigNode = OrigItem->Node;
	OrigItem->Node = NULL;
	return OrigNode;
}

void CCache::Store()
{
	DISArray<CCacheItem>::DIterator Itr = CacheArr.InitIterator();
	while (CacheArr.IsDereferenceable(Itr))
	{	CCacheItem *Current = CacheArr.GetCurrent(Itr);
		Tree->Store(Current->Node);
		++Itr;
	}
}

short CCache::Has(CSortedNode *Node)
{
	CObjPtr<CCacheItem> TestItem = new CCacheItem(Node, NOIDX, NOIDX);
	int FindIdx = CacheArr.Find(TestItem);
	TestItem->Node = NULL;
	return FindIdx != INT_MAX;
}

ushort CCache::FindIndex(long Offs)
{
	CSortedNode *TestNode = CNodeFactory::CreateTest(Offs, Tree);
	CObjPtr<CCacheItem> TestItem = new CCacheItem(TestNode, NOIDX, NOIDX);
	int FindIdx = CacheArr.Find(TestItem);
	return (ushort)(FindIdx == INT_MAX ? NOIDX : FindIdx);
}

void CCache::Del(long Offs)
{
	Assert(Offs != NOOFFS);
	ushort FindIdx = FindIndex(Offs);
	Assert(FindIdx != NOIDX);
	CCacheItem *Item = Detach(Offs, FindIdx);
	Tree->Store(Item->Node);
	delete Item;
}

CCacheItem *CCache::Detach(long Offs, ushort Idx)
{	CCacheItem *Item;

	// update ends of linked list
	if (Offs == FirstOffs)
		FirstOffs = CacheArr[Idx]->NextOffs;
	if (Offs == LastOffs)
		LastOffs = CacheArr[Idx]->PrevOffs;

	// update linked list
	if ((Item = GetPrev(Idx)) != NULL)
		SkipNext(Item);
	if ((Item = GetNext(Idx)) != NULL)
		SkipPrev(Item);

	Item = CacheArr[Idx];
	CacheArr.DetachByIndex(Idx);
#if defined(DEBUG_CCache)
	Check();
#endif
	return Item;
}

CCacheItem *CCache::GetPrev(ushort Idx)
{
	CCacheItem *Item = CacheArr[Idx];
	if (Item->PrevOffs == NOOFFS)
		return NULL;

	ushort PrevIdx = FindIndex(Item->PrevOffs);
	Assert(PrevIdx != NOIDX);
	return CacheArr[PrevIdx];
}

CCacheItem *CCache::GetNext(ushort Idx)
{
	CCacheItem *Item = CacheArr[Idx];
	if (Item->NextOffs == NOOFFS)
		return NULL;

	ushort NextIdx = FindIndex(Item->NextOffs);
	Assert(NextIdx != NOIDX);
	return CacheArr[NextIdx];
}

void CCache::SkipPrev(CCacheItem *Item)
{
	if (Item->PrevOffs == NOOFFS)
		return;

	ushort PrevIdx = FindIndex(Item->PrevOffs);
	Assert(PrevIdx != NOIDX);
	Item->PrevOffs = CacheArr[PrevIdx]->PrevOffs;
}

void CCache::SkipNext(CCacheItem *Item)
{
	if (Item->NextOffs == NOOFFS)
		return;

	ushort NextIdx = FindIndex(Item->NextOffs);
	Assert(NextIdx != NOIDX);
	Item->NextOffs = CacheArr[NextIdx]->NextOffs;
}

#if defined(DEBUG_CCache)
void CCache::Check()
{  CCacheItem *Item;
	ushort ChainCount;
	long Offs;
	ushort Idx;

	ushort Count = CacheArr.GetCount();

	ChainCount = 0;
	Offs = FirstOffs;
	while (Offs != NOOFFS)
	{	Idx = FindIndex(Offs);
		Assert(Idx != NOIDX && Idx < Count);
		Item = CacheArr[Idx];
		Offs = Item->NextOffs;
		ChainCount++;
		Assert(ChainCount <= Count);
	}
	Assert(ChainCount == Count);

	ChainCount = 0;
	Offs = LastOffs;
	while (Offs != NOOFFS)
	{	Idx = FindIndex(Offs);
		Assert(Idx != NOIDX && Idx < Count);
		Item = CacheArr[Idx];
		Offs = Item->PrevOffs;
		ChainCount++;
		Assert(ChainCount <= Count);
	}
	Assert(ChainCount == Count);
}
#endif

