#include "myfndt.h"
#include "tltree.h"
#include "lockptr.h"
#include "factory.h"
#include "handler.h"
#include "opendata.h"
#include <string.h>
#if defined(__GNUG__)
#include <unistd.h>
#endif

const int CTree::RNLimit = 20;
const int CTree::RNDelta = 10;

/* typedef */ struct SStreamHead
{	long FreeOffs;
	long RootOffs;
	long Reserved;
};

static void GetHead(CFileObj *Stream, SStreamHead &Head);

static void SetHead(CFileObj *Stream, SStreamHead &Head);

// file pointer must be 0!
// returns offset of root
static long InitHead(CFileObj *Stream);

// file pointer must be 0!
// returns offset of root
static long ReadHead(CFileObj *Stream);

static void GetHead(CFileObj *Stream, SStreamHead &Head)
{
	Stream->Seek(0);
	Stream->Read(&Head, sizeof(SStreamHead));
}

static void SetHead(CFileObj *Stream, SStreamHead &Head)
{
	Stream->Seek(0);
	Stream->Write(&Head, sizeof(SStreamHead));
}

static long InitHead(CFileObj *Stream)
{	SStreamHead head;

	head.FreeOffs = NOOFFS;
	head.RootOffs = sizeof(SStreamHead);
	head.Reserved = 0l;
	Stream->Write(&head, sizeof(SStreamHead));
	return head.RootOffs;
}

static long ReadHead(CFileObj *Stream)
{	SStreamHead head;

	GetHead(Stream, head);
	return head.RootOffs;
}

#if defined(_MSC_VER)
#pragma warning(disable:4355)
#endif

CTree::CTree(CKeyHandler *aKeyHandler,
	const DString &FName,
	ushort Opt):
	Stream(NULL),
	RemNodes(RNLimit, RNDelta),
	Cache(this),
	KeyHandler(aKeyHandler),
	Root(NULL),
	CreateOptions(Opt)
{
	if (OpenData(FName, Opt, StreamName, Stream))
		InitNew();
    else
	{   long offs = ReadHead(GetStream());
		Root = CNodeFactory::Create(offs, this, GetStream());
	}
}

#if defined(_MSC_VER)
#pragma warning(default:4355)
#endif

CTree::~CTree()
{
//	Assert((CreateOptions & IT_TEMP) || !Root->IsDirty());
		// not always - when an exception is thrown,
		// root can be dirty

	delete Root;
	if (CreateOptions & IT_TEMP)
	{	Stream.Close();
		unlink(StreamName.c_str());
	}
}

void CTree::InitNew()
{
	long offs = InitHead(GetStream());
	Root = CNodeFactory::Create(offs,
		NOOFFS,
		this,
		KeyHandler->GetAveKeyLen());
	Stream.Seek(offs);
	Stream.WriteZero(Root->GetBlockLen());
}

long CTree::AddOffs()
{	SStreamHead head;
	long offs, next;

	GetHead(GetStream(), head);
	if (head.FreeOffs == NOOFFS)
	{	Stream.Seek(0, SEEK_END);
   	offs = Stream.Tell();
		Stream.WriteZero(Root->GetBlockLen());
	}
	else
	{	Stream.Seek(head.FreeOffs);
		Stream.Read(&next, sizeof(long));
		offs = head.FreeOffs;
		head.FreeOffs = next;
		SetHead(GetStream(), head);
	}
	return offs;
}

void CTree::DelOffs(long Offs)
{	SStreamHead head;

	GetHead(GetStream(), head);
	Stream.Seek(Offs);
	Stream.Write(&(head.FreeOffs), sizeof(long));
	head.FreeOffs = Offs;
	SetHead(GetStream(), head);
}

CNode *CTree::GetLockedNode(long Offs)
{	CSortedNode *FoundNode;

#if defined(DEBUG_duplicate)
	CheckDupl();
#endif

	if (Root->GetOffs() == Offs)
		FoundNode = Root;
	else
	{
		CObjPtr<CSortedNode> TestNode = CNodeFactory::CreateTest(Offs, this);
		int RemIdx = RemNodes.Find(TestNode);
		if (RemIdx != INT_MAX) // found
			FoundNode = RemNodes[RemIdx];
		else
		{	if ((FoundNode = Cache.Yank(Offs)) == NULL)
				FoundNode = CNodeFactory::Create(Offs, this, GetStream());
			RemNodes.Add(FoundNode);
		}
	}

	FoundNode->IncLockCount();

#if defined(DEBUG_duplicate)
	CheckDupl();
	Assert(IsLocked(FoundNode));
#endif

	return FoundNode;
}

short CTree::UnlockNode(CNode *Node)
{
#if defined(DEBUG_duplicate)
	Assert(IsLocked(Node));
#endif
	short forgotten = ((CSortedNode *)Node)->DecLockCount();
	if (forgotten) // node no longer locked
	{  if (Root->GetOffs() != Node->GetOffs())
		{	Assert(RemNodes.Find((CSortedNode *)Node) != INT_MAX);
			RemNodes.Detach((CSortedNode *)Node);
			Remember(Node);
		}
		// else nothing - root stays in memory even when not locked
	}
	return forgotten;
}

void CTree::AddLevel(CItem *Item)
{	long Offs, OldRootOffs;
	SStreamHead head;

#if defined(DEBUG_duplicate)
	CheckDupl();
#endif

	CObjPtr<CItem> ItemPtr = Item;
	Offs = AddOffs();
	OldRootOffs = Root->GetOffs();

	ItemPtr->UpdateRef(Offs, this);
		// parent of item's reference is the new root

	CItem **Info = AllocInfo();
	Info[0] = ItemPtr.Drop();
	CObjPtr<CSortedNode> NewRoot = CNodeFactory::Create(Offs,
		NOOFFS,
		this,
		Info, // reference will be updated
		1,
		OldRootOffs, // will be updated
		GetKeyHandler()->GetAveKeyLen());

	Root->SetParent(Offs);
	if (Root->IsLocked())
	{	Assert(RemNodes.Find(Root) == INT_MAX);
		RemNodes.Add(Root);
	}
	else
	{	Store(Root);
		delete Root;
	}
	Root = NewRoot.Drop();

	// save new root offset into SStreamHead
	GetHead(GetStream(), head);
	head.RootOffs = Root->GetOffs();
	SetHead(GetStream(), head);

#if defined(DEBUG_duplicate)
	CheckDupl();
#endif
}

void CTree::CheckDelLevel(CNode *NewRoot)
{	SStreamHead head;

	Assert(((CSortedNode *)NewRoot)->IsLocked());
	if (!Root->IsEmpty())
	{	UnlockNode(NewRoot);
		return;
	}

	UnlockNode(NewRoot); // may put it into cache
	Cache.Yank(NewRoot->GetOffs());
	NewRoot->SetParent(NOOFFS);
	DestroyUnlocked(Root);
	Root = (CSortedNode *)NewRoot;

	// save new root offset into SStreamHead
	GetHead(GetStream(), head);
	head.RootOffs = Root->GetOffs();
	SetHead(GetStream(), head);
}

void CTree::DestroyNode(CNode *Doomed)
{
	// Root must be destroyed only by CTree using DestroyUnlocked,
	// not by this public function
	Assert(!Doomed->IsRoot());

	// Doomed must be locked exactly once - by the node which
	// wants to destroy it
	Assert(((CSortedNode *)Doomed)->IsLocked());
	UnlockNode(Doomed); // moves it into cache; inefficient?
	Assert(!((CSortedNode *)Doomed)->IsLocked());

	DestroyUnlocked(Doomed);
}

void CTree::DestroyUnlocked(CNode *Doomed)
{
	Assert(!((CSortedNode *)Doomed)->IsLocked());
#if defined(DEBUG_duplicate)
	CheckDupl();
#endif

	long Offs = Doomed->GetOffs();
	Cache.Yank(Offs);
	delete Doomed;
	DelOffs(Offs);
}

// does not implement error recovery yet
void CTree::Store(CNode *Node)
{
	if (Node->IsDirty())
		Node->Save();
}

void CTree::Remember(CNode *Node)
{
#if defined(DEBUG_duplicate)
	CheckDupl();
#endif

	if (Node->GetOffs() != Root->GetOffs())
		Cache.Add((CSortedNode *)Node);
}

void CTree::StoreAll()
{
	Assert(!(CreateOptions & IT_TEMP));

	Store(Root);

	// store RemNodes
	DISArray<CSortedNode>::DIterator Itr = RemNodes.InitIterator();
	while (RemNodes.IsDereferenceable(Itr))
	{	CSortedNode *Current = RemNodes.GetCurrent(Itr);
		Store(Current);
		++Itr;
	}

	Cache.Store();
}

void CTree::UpdateNode(long Offs, long Parent)
{
	if (Offs != NOOFFS)
	{	CLockedPtr Node(GetLockedNode(Offs), this);
		Node->SetParent(Parent);
	}
}

#if defined(DEBUG_duplicate)
void CTree::CheckDupl()
{
	Assert(!Cache.Has(Root));
	ushort Count = RemNodes.GetItemsInContainer();
	for (ushort i = 0; i < Count; i++)
		Assert(!Cache.Has(RemNodes[i]));
}

int CTree::IsLocked(CNode *Node)
{
	CheckRemNodes();
	if (Node->GetOffs() == Root->GetOffs())
		return 1;
	int RemIdx = RemNodes.Find((CSortedNode *)Node);
	return RemIdx != INT_MAX;
}

void CTree::CheckRemNodes()
{	CSortedNode *Node;

	ushort Count = RemNodes.GetItemsInContainer();
	for (ushort i = 0; i < Count; i++)
	{  Node = RemNodes[i];
		Assert(Node->IsLocked());
	}
}
#endif

